import pandas as pd
from tqdm import tqdm
from .pull_cloudtrail_lookupEvents import GeneralManager as GraCloudtrailManager
import os
from ..utils import NoCloudtrailException

import logging
logger = logging.getLogger('isitfit')



class Manager:
    def __init__(self, EndTime):
        self.EndTime = EndTime

    def init_data(self, context_pre):
        # parse out of context
        ec2_instances, region_include, n_ec2 = context_pre['ec2_instances'], context_pre['region_include'], context_pre['n_ec2_total']

        # get cloudtail ec2 type changes for all instances
        self.region_include = region_include
        self.df_cloudtrail = self._fetch()

        # first pass to append ec2 types to cloudtrail based on "now"
        self.df_cloudtrail = self.df_cloudtrail.reset_index()
        # Edit 2019-11-12 use initial=0 otherwise if "=1" used then the tqdm output would be "101it" at conclusion, i.e.
        # First pass through EC2 instances: 101it [00:05,  5.19it/s]
        for ec2_dict, ec2_id, ec2_launchtime, ec2_obj in tqdm(ec2_instances, total=n_ec2, desc="Pass 1/2 through EC2 instances", initial=0):
            self._appendNow(ec2_obj)

        # set index again, and sort decreasing this time (not like git-remote-aws default)
        self.df_cloudtrail = self.df_cloudtrail.set_index(["instanceId", "EventTime"]).sort_index(ascending=False)

        # done
        return context_pre


    def _fetch(self):
        # get cloudtrail ec2 type changes for all instances
        logger.debug("Downloading cloudtrail data (from %i regions)"%len(self.region_include))
        df_2 = []
        import boto3
        cloudtrail_client_all = {}
        for region_name in self.region_include:
          if region_name not in cloudtrail_client_all.keys():
            boto3.setup_default_session(region_name = region_name)
            cloudtrail_client_all[region_name] = boto3.client('cloudtrail')

          cloudtrail_client_single = cloudtrail_client_all[region_name]
          cloudtrail_manager = GraCloudtrailManager(cloudtrail_client_single)
          df_1 = cloudtrail_manager.ec2_typeChanges()
          df_2.append(df_1)

        # concatenate
        df_3 = pd.concat(df_2, axis=0)
        return df_3

    """
    # Cached version ... disabled because not sure how to generalize it
    def _fetch(self):
        # get cloudtail ec2 type changes for all instances
        # FIXME
        cache_fn = '/tmp/isitfit_cloudtrail.shadiakiki1986.csv'
        # cache_fn = '/tmp/isitfit_cloudtrail.autofitcloud.csv'
        if os.path.exists(cache_fn):
            logger.debug("Loading cloudtrail data from cache")
            df = pd.read_csv(cache_fn).set_index(["instanceId", "EventTime"])
            return df

        logger.debug("Downloading cloudtrail data")
        cloudtrail_manager = GraCloudtrailManager(cloudtrail_client)
        df = cloudtrail_manager.ec2_typeChanges()

        # save to cache
        df.to_csv(cache_fn)

        # done
        return df
    """


    def _appendNow(self, ec2_obj):
        # artificially append an entry for "now" with the current type
        # This is useful for instance who have no entries in the cloudtrail
        # so that their type still shows up on merge
        df_new = pd.DataFrame([
              { 'instanceId': ec2_obj.instance_id,
                'EventTime': self.EndTime,
                'instanceType': ec2_obj.instance_type
              }
            ])

        self.df_cloudtrail = pd.concat([self.df_cloudtrail, df_new], sort=True)


    def single(self, context_ec2):
        ec2_obj = context_ec2['ec2_obj']

        # pandas series of number of cpu's available on the machine over time, past 90 days
        # series_type_ts1 = self.cloudtrail_client.get_ec2_type(ec2_obj.instance_id)
        if not ec2_obj.instance_id in self.df_cloudtrail.index:
            return None

        df_type_ts1 = self.df_cloudtrail.loc[ec2_obj.instance_id]
        if df_type_ts1 is None:
          raise NoCloudtrailException("No cloudtrail data for %s"%ec2_obj.instance_id)

        # set in context
        context_ec2['df_type_ts1'] = df_type_ts1

        # done
        return context_ec2


class CloudtrailCached(Manager):
    def __init__(self, EndTime, cache_man):
        super().__init__(EndTime)
        self.cache_man = cache_man


    def _fetch(self):
        # get cloudtrail ec2 type changes for all instances

        # check cache first
        cache_key = "cloudtrail_ec2type._fetch"
        if self.cache_man.isReady():
          df_cache = self.cache_man.get(cache_key)
          if df_cache is not None:
            logger.debug("Found cloudtrail data in redis cache")
            return df_cache

        # if no cache, then download
        df_fresh = super()._fetch()

        # if caching enabled, store it for later fetching
        # https://stackoverflow.com/a/57986261/4126114
        if self.cache_man.isReady():
          self.cache_man.set(cache_key, df_fresh)

        # done
        return df_fresh




# 2019-10-24 Not sure why find_packages was commented out .. bringing it back
from setuptools import setup, find_packages

# copied from https://github.com/awslabs/git-remote-codecommit/blob/master/setup.py
import os
def read(fname):
  return open(os.path.join(os.path.dirname(__file__), fname)).read()
  

from isitfit import isitfit_version

# follow https://github.com/awslabs/git-remote-codecommit/blob/master/setup.py
# and https://packaging.python.org/tutorials/packaging-projects/
setup(
    name='isitfit',
    version=isitfit_version,
    author="Shadi Akiki, AutofitCloud",
    author_email="shadi@autofitcloud.com",
    url='https://gitlab.com/autofitcloud/isitfit',
    description="Command-line tool to calculate excess AWS cloud resource capacity",

    # 2019-09-10 not sure what in the README.md is yielding the twine error
    # The description failed to render in the default format of reStructuredText.
    # long_description = read('README.md'),
    long_description = 'Check https://isitfit.autofitcloud.com',
    long_description_content_type="text/markdown",
    
    packages=find_packages(),
    # packages = ['isitfit'],
    include_package_data=True,
    install_requires=[
        'click==7.0',
        'pandas==0.25.1',
        'requests==2.22.0',
        'cachecontrol==0.12.5',
        'lockfile==0.12.2',
        'tabulate==0.8.3',
        'termcolor==1.1.0',
        'tqdm==4.32.2',

        'redis==3.3.8',
        'pyarrow==0.15.0',

        # check note in requirements.txt
        'awscli==1.16.248',
        'boto3==1.9.238',

        'datadog==0.30.0',
        'schema==0.7.1',
        'visidata==1.5.2',
        'outdated==0.2.0',
        'aws-requests-auth==0.4.2',
        'matomo_sdk_py==0.1.0',
        'simple-cache==0.35'
    ],
    entry_points='''
        [console_scripts]
        isitfit=isitfit.cli.core:cli_core
    ''',
)



import pandas as pd

import logging
logger = logging.getLogger('isitfit')

class TagsImplierHelper:
  def __init__(self, names_df):
    self.names_df = names_df
    self.names_original = names_df.Name.tolist()

  def freq_list(self):
    logger.info("Step 1: calculate word frequencies")
    # lower-case
    self.names_lower = [x.lower() for x in self.names_original]

    # count single word frequencies
    # https://programminghistorian.org/en/lessons/counting-frequencies
    import re
    #names_split = (' '.join(names_original)).split(' ')
    #words = re.findall("\w+", "the quick person did not realize his speed and the quick person bumped")
    names_split = re.findall("\w+", ' '.join(self.names_lower))

    # Counting bigrams
    # https://stackoverflow.com/a/12488794/4126114
    from itertools import tee, islice
    from collections import Counter
    def ngrams(lst, n):
      tlst = lst
      while True:
        a, b = tee(tlst)
        l = tuple(islice(a, n))
        if len(l) == n:
          yield l
          next(b)
          tlst = b
        else:
          break

    def get_freq(n):
      # names_freq = dict(Counter(zip(names_split, islice(names_split, n-1, None))))
      names_freq = dict(Counter(ngrams(names_split, n)))
      names_freq = [(k,v) for k,v in names_freq.items()]
      return names_freq

    self.freq_1w = get_freq(1)
    self.freq_2w = get_freq(2)

  def freq_df(self):
    logger.info("Step 2: convert word frequencies to pandas dataframe")

    # convert to pandas dataframe
    
    def min_is_2(df_in):
      return df_in[df_in.n >= 2] # minimum occurence is 2
    
    def freq2df(freq_in, l):
      df_freq_in = pd.DataFrame(freq_in, columns=['word_tuple', 'n'])
      df_freq_in = min_is_2(df_freq_in)
      df_freq_in['l'] = l
      return df_freq_in
    
    df_freq_1w = freq2df(self.freq_1w, 1)
    df_freq_1w['word_1'] = df_freq_1w.word_tuple.apply(lambda x: x[0])
    df_freq_1w['word_2'] = None
    df_freq_2w = freq2df(self.freq_2w, 2)
    df_freq_2w['word_1'] = df_freq_2w.word_tuple.apply(lambda x: x[0])
    df_freq_2w['word_2'] = df_freq_2w.word_tuple.apply(lambda x: x[1])
    
    # print("##########")
    # print("before filter")
    # print("1w")
    # print(df_freq_1w)
    # print("")
    # print("2w")
    # print(df_freq_2w)
    # print("##########")
    
    
    # filter out 1-grams if their 2-gram counterpart is superior
    df_freq_2w = df_freq_2w.merge(df_freq_1w[['word_1', 'n']], how='left', left_on='word_1', right_on='word_1', suffixes=['', '.1w=2w.word1'])
    df_freq_2w = df_freq_2w.merge(df_freq_1w[['word_1', 'n']], how='left', left_on='word_2', right_on='word_1', suffixes=['', '.1w=2w.word2'])
    df_freq_2w = df_freq_2w.drop(columns=['word_1.1w=2w.word2'])
    df_freq_2w = df_freq_2w[(df_freq_2w.n >= df_freq_2w['n.1w=2w.word1']) & (df_freq_2w.n >= df_freq_2w['n.1w=2w.word2'])]
    
    # print("")
    # print("after filtering 2w")
    # print(df_freq_2w)
    
    # drop from 1w the components that were used in the 2w
    df_freq_1w = df_freq_1w[~(df_freq_1w.word_1.isin(df_freq_2w.word_1) | df_freq_1w.word_1.isin(df_freq_2w.word_2))]
    
    # drop columns
    df_freq_2w = df_freq_2w.drop(columns=['n.1w=2w.word1', 'n.1w=2w.word2'])
    
    # concatenate into 1 df
    df_freq_all = pd.concat([df_freq_1w,df_freq_2w], axis=0)
    df_freq_all['word_combined'] = df_freq_all.apply(lambda r: r.word_1 if r.word_2 is None else r.word_1 + ' ' + r.word_2, axis=1)
    
    #print("")
    #print("final df")
    #print(df_freq_all)
    self.df_freq_all = df_freq_all

  def tag_set(self):
    logger.info("Step 3: Generate a set of tags")

    # merge with original names
    df_ori = pd.DataFrame({'original': self.names_original, 'lower': self.names_lower, 'instance_id': self.names_df.instance_id})
    df_ori['tag_set'] = None
    
    import re
    def myfind(t, n):
      # https://stackoverflow.com/a/48205793/4126114
      f = re.findall(r"\b%s\b"%t, n, re.IGNORECASE)
      return len(f)>0

    self.len_ori = df_ori.shape[0]
    for i in range(self.len_ori):
      name_value = df_ori.iloc[i].lower
      for j in range(self.df_freq_all.shape[0]):
        i_sub = [myfind(tag_value, name_value) for tag_value in self.df_freq_all.word_combined.values]
        df_sub = self.df_freq_all[i_sub]
        df_sub = df_sub.sort_values('n', ascending=False)
        df_sub = df_sub.iloc[:3]
        df_ori.at[i, 'tag_set'] = set(df_sub.word_combined.values)
    
    self.df_ori = df_ori

  def tag_list(self):
    logger.info("Step 4: convert the set of tags to a list of tags")

    df_ori = self.df_ori

    # initialize
    # just doing [[None]*3]*len_ori doesn't work
    df_ori['tag_list'] = None
    for i1 in range(self.len_ori):
      df_ori.at[i1, 'tag_list'] = [None]*3
    
    # distributing the tag_set to tag_1, tag_2, tag_3 in such a way that for example "app" is at tag_1 for all the instances
    tag_processed = set()
    for i1 in range(self.len_ori):
      for tag_value in df_ori.iloc[i1].tag_set:
        if tag_value in tag_processed:
          continue
    
        tag_processed.add(tag_value)
        logger.debug("<<<<<<<<>>>>>>>>>>>>")
        logger.debug("%i: %s"%(i1, tag_value))
        logger.debug(df_ori)
    
        if tag_value in df_ori.at[i1, 'tag_list']:
          continue # already inserted this tag
    
        # find free indeces in current list
        if None not in df_ori.at[i1, 'tag_list']:
          raise Exception("No more space in list for %s"%tag_value)
    
        # https://stackoverflow.com/a/6294205/4126114
        free_indices = [i for i, x in enumerate(df_ori.at[i1, 'tag_list']) if x is None]

        # find the first free index which is ok for all entries having this tag
        free_chosen = None
        logger.debug("Searching for free index for %s"%tag_value)
        for free_i1 in free_indices:
          found_conflict = False
          for i2 in range(self.len_ori):
            if found_conflict: break
            if i2 <= i1: continue
            logger.debug("Checking row %i"%i2)
            # if tag in set of tags for this 2nd row
            if tag_value in df_ori.loc[i2].tag_set:
              # and if the value for this tag is not *already* set
              if tag_value not in df_ori.loc[i2].tag_list:
                if df_ori.loc[i2, 'tag_list'][free_i1] is not None:
                  logger.debug("Found conflict")
                  found_conflict = True

          if not found_conflict:
            logger.debug("Found chosen free index at %i"%free_i1)
            free_chosen = free_i1
            break

        # if no free index chosen, raise Exception
        if free_chosen is None:
          raise Exception("Conflict found: %s didn't find a free index to use"%(tag_value))

        # otherwise use the chosen index
        # Old way of getting first None only # free_chosen = df_ori.at[i1, 'tag_list'].index(None)
        free_chosen = free_i1
        df_ori.at[i1, 'tag_list'][free_chosen] = tag_value
    
        # set this tag for all other rows at "free_chosen"
        for i2 in range(self.len_ori):
          if i2 <= i1: continue
          if tag_value in df_ori.loc[i2].tag_set:
            if tag_value not in df_ori.loc[i2].tag_list:
              if df_ori.loc[i2, 'tag_list'][free_chosen] is not None:
                raise Exception("Conflict found despite pre-check? %s wants to be at %i but found %s already"%(tag_value, free_chosen, df_ori.loc[i2, 'tag_list'][free_chosen]))
    
            df_ori.at[i2, 'tag_list'][free_chosen] = tag_value
    
    # mesh out the tag_list to tag_1 tag_2 tag_3
    df_ori['tag_1'] = df_ori.tag_list.apply(lambda x: x[1-1])
    df_ori['tag_2'] = df_ori.tag_list.apply(lambda x: x[2-1])
    df_ori['tag_3'] = df_ori.tag_list.apply(lambda x: x[3-1])

    # re-order columns
    df_ori = df_ori.rename(columns={'original': 'instance_name'})
    df_ori = df_ori[['instance_id', 'instance_name', 'tag_1', 'tag_2', 'tag_3']]
    
    # done
    #print("")
    #print("tagged")
    #print(df_ori)

    self.df_ori = df_ori

#------------------------

class TagsImplierMain:

  def __init__(self, names_original):
    self.helper = TagsImplierHelper(names_original)

  def imply(self):
    self.helper.freq_list()
    self.helper.freq_df()
    self.helper.tag_set()
    self.helper.tag_list()
    return self.helper.df_ori

#------------------------

if __name__=='__main__':
    names_original = [
      'cool php app for accounting',
      'weird app for BUSINESS',
      'a crm portal for accounting',
      'crm portal for business',
    ]
    
    #tags_expected = [
    #  ['app', 'accounting', 'for'],
    #  ['app', 'business',   'for'],
    #  ['crm portal', 'accounting', 'for'],
    #  ['crm portal', 'business',   'for'],
    #]
    
    tags_implier = TagsImplierMain(names_original)
    df_tagged = tags_implier.imply()
    print(df_tagged)



import logging
logger = logging.getLogger('isitfit')

from ..utils import SECONDS_IN_ONE_DAY
SECONDS_IN_10MINS = 60*10

class RedisPandas:
  """
  Python class that manages caching pandas dataframes to redis
  https://stackoverflow.com/a/57986261/4126114
  """
  def __init__(self):
    self.redis_args = {}
    self.redis_client = None
    self.pyarrow_context = None

  def fetch_envvars(self):
    # check redis parameters if set for caching
    import os
    for k1, k2 in [
        ('host', "ISITFIT_REDIS_HOST"),
        ('port', "ISITFIT_REDIS_PORT"),
        ('db', "ISITFIT_REDIS_DB")
      ]:
      self.redis_args[k1] = os.getenv(k2, None)

  def isSetup(self):
    return all(self.redis_args.values())

  def connect(self):
    logger.info("Connecting to redis cache")
    logger.debug(self.redis_args)
    import redis
    import pyarrow as pa

    self.redis_client = redis.Redis(**self.redis_args)
    self.pyarrow_context = pa.default_serialization_context()

  def isReady(self):
    return self.redis_client is not None

  def set(self, key, df):
    # Note that in case data was not found, eg in mainManager._cloudwatch_metrics_core, an empty dataframe is returned (and thus passed in here)
    pybytes = self.pyarrow_context.serialize(df).to_buffer().to_pybytes()
    # set expiration of key-value pair to be 1 day if data was found, 10 minutes otherwise
    ex = SECONDS_IN_10MINS if df.shape[0]==0 else SECONDS_IN_ONE_DAY
    # https://redis-py.readthedocs.io/en/latest/#redis.Redis.set
    self.redis_client.set(name=key, value=pybytes, ex=ex)

  def get(self, key):
    v1 = self.redis_client.get(key)
    if not v1: return v1
    v2 = self.pyarrow_context.deserialize(v1)
    return v2

  def handle_pre(self, context_pre):
        # set up caching if requested
        self.fetch_envvars()
        if self.isSetup():
          self.connect()

        # 0th pass to count
        n_ec2_total = context_pre['n_ec2_total']

        # if more than 10 servers, recommend caching with redis
        if n_ec2_total > 10 and not self.isSetup():
            from termcolor import colored
            logger.warning(colored(
"""Since the number of EC2 instances is %i,
it is recommended to use redis for caching of downloaded CPU/memory metrics.
To do so
- install redis

    [sudo] apt-get install redis-server

- export environment variables

    export ISITFIT_REDIS_HOST=localhost
    export ISITFIT_REDIS_PORT=6379
    export ISITFIT_REDIS_DB=0

where ISITFIT_REDIS_DB is the ID of an unused database in redis.

And finally re-run isitfit as usual.
"""%n_ec2_total, "yellow"))
            continue_wo_redis = input(colored('Would you like to continue without redis caching (not recommended)? yes/[no] ', 'cyan'))
            if not (continue_wo_redis.lower() == 'yes' or continue_wo_redis.lower() == 'y'):
                from isitfit.utils import IsitfitCliError
                raise IsitfitCliError("Aborting to set up redis.", context_pre['click_ctx'])


        # done
        return context_pre



# Mostly a copy of test_iterator_redshift
# Need to convert the latter to class
# and inherit here to avoid code redundancy


from ....cost.redshift.cloudwatchman import CloudwatchEc2

import pytest
from isitfit.utils import NoCloudwatchException

def test_init():
  rpi = CloudwatchEc2()
  assert True # no exception

# cannot use mock_cloudwatch
# yields error:
# botocore.exceptions.PaginationError: Error during pagination: The same next token was received twice
#from moto import mock_cloudwatch
#@mock_cloudwatch
def test_handleCluster_notFound(mocker):
  mockreturn = lambda *args, **kwargs: []
  mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchEc2._metrics_filter'
  mocker.patch(mockee, side_effect=mockreturn)

  rpi = CloudwatchEc2()
  dummy_id = 'abc'
  with pytest.raises(NoCloudwatchException):
    m_i = rpi.handle_cluster(dummy_id)


def test_handleCluster_foundCluster(mocker):
  class MockMetricCluster:
    dimensions = [1]

  mockreturn = lambda *args, **kwargs: [MockMetricCluster]
  mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchEc2._metrics_filter'
  mocker.patch(mockee, side_effect=mockreturn)

  rpi = CloudwatchEc2()
  dummy_id = 'abc'
  m_i = rpi.handle_cluster(dummy_id)
  assert m_i is not None


def test_handleCluster_foundMany(mocker):
  class MockMetricCluster:
    dimensions = [1]

  class MockMetricNode:
    dimensions = [1, 2]

  mockreturn = lambda *args, **kwargs: [MockMetricNode, MockMetricCluster]
  mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchEc2._metrics_filter'
  mocker.patch(mockee, side_effect=mockreturn)

  rpi = CloudwatchEc2()
  dummy_id = 'abc'
  m_i = rpi.handle_cluster(dummy_id)
  assert m_i is not None


def test_handleMetric_empty(mocker):
  mockreturn = lambda *args, **kwargs: {'Datapoints': []}
  mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchEc2._metric_get_statistics'
  mocker.patch(mockee, side_effect=mockreturn)

  rpi = CloudwatchEc2()
  with pytest.raises(NoCloudwatchException):
    df = rpi.handle_metric(None, None, None)


def test_handleMetric_notEmpty(mocker):
  import datetime as dt
  dt_now = dt.datetime.utcnow()

  ex_dp = [
    {'Timestamp': dt_now - dt.timedelta(seconds=1)},
    {'Timestamp': dt_now - dt.timedelta(seconds=2)},
    {'Timestamp': dt_now - dt.timedelta(seconds=3)}
  ]
  mockreturn = lambda *args, **kwargs: {'Datapoints': ex_dp}
  mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchEc2._metric_get_statistics'
  mocker.patch(mockee, side_effect=mockreturn)

  rpi = CloudwatchEc2()
  df = rpi.handle_metric(None, None, dt_now)
  assert df is not None



import datetime as dt
dt_now_d = dt.datetime.utcnow()


import pytest
@pytest.fixture(scope='function')
def MockCwResource(mocker):
    class MockMetricsObj:
        dimensions = ['one item']
        def get_statistics(self, *args, **kwargs):
            return {'Datapoints': [{'Timestamp':dt.datetime.now(), 'b':2}]}

    class MockMetricsIterator:
        n = 1
        def filter(self, *args, **kwargs):
            # yield 1 object is ok
            # yield >1 objects triggers exception
            for i in range(self.n):
                yield MockMetricsObj()

    class MyCwResource:
      metrics = MockMetricsIterator()

    return MyCwResource


from moto import mock_ec2, mock_cloudwatch # TODO not in moto:, mock_cloudtrail

class TestCloudwatchEc2:
# Edit 2019-11-18
# After starting to use isitfit.cost.redshift.cloudwatchman in mainManager,
# the checks for multiple metrics was dropped as it was useless really.
# Now it just returns the first non-empty entry.
#  @mock_ec2
#  @mock_cloudwatch
#  # @mock_cloudtrail
#  def test_perEc2_failMultiple(self, MockCwResource, mocker):
#    from ...cost.mainManager import MainManager
#    mm = MainManager(None, None)
#
#    # mock resource
#    mcw = MockCwResource()
#    mcw.metrics.n = 2 # set to 2 to trigger exception
#    mockreturn = lambda *args, **kwargs: mcw
#    mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchBase._cloudwatch_metrics_boto3'
#    mocker.patch(mockee, side_effect=mockreturn)
#
#    # class for ec2_obj
#    class MockEc2Obj:
#      region_name = 'us-west-2'
#      instance_id = 'i1'
#      launch_time = dt_now_d
#
#    ec2_obj = MockEc2Obj()
#
#    import pytest
#    from ...utils import IsitfitCliError
#    with pytest.raises(IsitfitCliError) as e:
#      # raise exception
#      mm._cloudwatch_metrics_cached(ec2_obj)



  @mock_ec2
  @mock_cloudwatch
  # @mock_cloudtrail
  def test_perEc2_ok(self, MockCwResource, mocker):
    from isitfit.cost.redshift.cloudwatchman import CloudwatchEc2
    cw = CloudwatchEc2(None)

    # mock resource
    mcw = MockCwResource()
    mcw.metrics.n = 1 # set to 1 to NOT trigger exception
    mockreturn = lambda *args, **kwargs: mcw
    mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchBase._cloudwatch_metrics_boto3'
    mocker.patch(mockee, side_effect=mockreturn)

    # class for ec2_obj
    class MockEc2Obj:
      region_name = 'us-west-2'
      instance_id = 'i1'
      launch_time = dt_now_d

    ec2_obj = MockEc2Obj()

    cw.per_ec2({'ec2_obj': ec2_obj})
    assert True # no exception

